use chrono::prelude::*;
use diesel::insert_into;
use diesel::prelude::*;
use diesel::update;

use crate::schema::{clarification, contest_problems, user};
use crate::user::{User, USER_COLUMNS};

#[derive(Queryable)]
pub struct Clarification {
    pub uuid: String,
    pub question: String,
    pub question_instant: NaiveDateTime,
    pub answer: Option<String>,
    pub answer_instant: Option<NaiveDateTime>,
    pub answer_user_id: Option<i32>,
    pub answer_to_all: bool,
    pub contest_problem_id: Option<i32>,
    pub contest_id: i32,
    pub user_id: i32,
}

pub const CLARIFICATION_COLUMNS: (
    clarification::uuid,
    clarification::question,
    clarification::question_instant,
    clarification::answer,
    clarification::answer_instant,
    clarification::answer_user_id,
    clarification::answer_to_all,
    clarification::contest_problem_id,
    clarification::contest_id,
    clarification::user_id,
) = (
    clarification::uuid,
    clarification::question,
    clarification::question_instant,
    clarification::answer,
    clarification::answer_instant,
    clarification::answer_user_id,
    clarification::answer_to_all,
    clarification::contest_problem_id,
    clarification::contest_id,
    clarification::user_id,
);

#[derive(Insertable)]
#[diesel(table_name = clarification)]
pub struct NewClarification {
    pub uuid: String,
    pub question: String,
    pub question_instant: NaiveDateTime,
    pub answer_to_all: bool,
    pub contest_problem_id: Option<i32>,
    pub contest_id: i32,
    pub user_id: i32,
}

pub fn insert_clarification(
    connection: &mut PgConnection,
    new_clarification: NewClarification,
) -> QueryResult<Clarification> {
    let uuid = new_clarification.uuid.clone();
    insert_into(clarification::table)
        .values(new_clarification)
        .execute(connection)?;
    clarification::table
        .filter(clarification::uuid.eq(uuid))
        .first(connection)
}

pub fn get_clarifications_by_contest_user(
    connection: &mut PgConnection,
    user_id: i32,
    contest_id: i32,
) -> QueryResult<Vec<(Clarification, User, Option<String>)>> {
    clarification::table
        .inner_join(user::table)
        .left_join(contest_problems::table)
        .filter(
            clarification::answer_to_all
                .eq(true)
                .or(clarification::user_id.eq(user_id)),
        )
        .filter(clarification::contest_id.eq(contest_id))
        .select((
            CLARIFICATION_COLUMNS,
            USER_COLUMNS,
            contest_problems::label.nullable(),
        ))
        .order(clarification::question_instant.desc())
        .load(connection)
}

pub fn get_clarifications_by_contest(
    connection: &mut PgConnection,
    contest_id: i32,
) -> QueryResult<Vec<(Clarification, User, Option<String>)>> {
    clarification::table
        .inner_join(user::table)
        .left_join(contest_problems::table)
        .filter(clarification::contest_id.eq(contest_id))
        .select((
            CLARIFICATION_COLUMNS,
            USER_COLUMNS,
            contest_problems::label.nullable(),
        ))
        .order(clarification::question_instant.desc())
        .load(connection)
}

pub fn update_clarification(
    connection: &mut PgConnection,
    uuid: String,
    new_answer: String,
    new_answer_to_all: bool,
    new_answer_user_id: i32,
    new_answer_instant: NaiveDateTime,
) -> QueryResult<()> {
    update(clarification::table)
        .filter(clarification::uuid.eq(uuid))
        .set((
            clarification::answer.eq(new_answer),
            clarification::answer_to_all.eq(new_answer_to_all),
            clarification::answer_user_id.eq(new_answer_user_id),
            clarification::answer_instant.eq(new_answer_instant),
        ))
        .execute(connection)?;
    Ok(())
}
