use chrono::prelude::*;
use diesel::insert_into;
use diesel::prelude::*;
use diesel::update;

use crate::schema::{printing_task, user};
use crate::user::{User, USER_COLUMNS};

#[derive(Queryable)]
pub struct PrintingTask {
    pub uuid: String,
    pub contents: String,
    pub creation_instant: NaiveDateTime,
    pub status: String,
    pub done_instant: Option<NaiveDateTime>,
    pub contest_id: i32,
    pub user_id: i32,
}

pub const PRINTING_TASK_COLUMNS: (
    printing_task::uuid,
    printing_task::contents,
    printing_task::creation_instant,
    printing_task::status,
    printing_task::done_instant,
    printing_task::contest_id,
    printing_task::user_id,
) = (
    printing_task::uuid,
    printing_task::contents,
    printing_task::creation_instant,
    printing_task::status,
    printing_task::done_instant,
    printing_task::contest_id,
    printing_task::user_id,
);

#[derive(Insertable)]
#[diesel(table_name = printing_task)]
pub struct NewPrintingTask {
    pub uuid: String,
    pub contents: String,
    pub creation_instant: NaiveDateTime,
    pub status: String,
    pub done_instant: Option<NaiveDateTime>,
    pub contest_id: i32,
    pub user_id: i32,
}

pub fn insert_printing_task(
    connection: &mut PgConnection,
    new_printing_task: NewPrintingTask,
) -> QueryResult<PrintingTask> {
    let uuid = new_printing_task.uuid.clone();
    insert_into(printing_task::table)
        .values(new_printing_task)
        .execute(connection)?;
    printing_task::table
        .filter(printing_task::uuid.eq(uuid))
        .first(connection)
}

pub fn get_printing_tasks_by_contest_user(
    connection: &mut PgConnection,
    user_id: i32,
    contest_id: i32,
) -> QueryResult<Vec<PrintingTask>> {
    printing_task::table
        .filter(printing_task::user_id.eq(user_id))
        .filter(printing_task::contest_id.eq(contest_id))
        .load(connection)
}

pub fn get_printing_tasks_by_contest(
    connection: &mut PgConnection,
    contest_id: i32,
) -> QueryResult<Vec<(PrintingTask, User)>> {
    printing_task::table
        .inner_join(user::table)
        .filter(printing_task::contest_id.eq(contest_id))
        .select((PRINTING_TASK_COLUMNS, USER_COLUMNS))
        .load(connection)
}

pub fn get_printing_task_by_uuid(
    connection: &mut PgConnection,
    uuid: String,
) -> QueryResult<(PrintingTask, User)> {
    printing_task::table
        .inner_join(user::table)
        .filter(printing_task::uuid.eq(uuid))
        .select((PRINTING_TASK_COLUMNS, USER_COLUMNS))
        .first(connection)
}

pub fn update_printing_task(
    connection: &mut PgConnection,
    uuid: String,
    new_done_instant: NaiveDateTime,
) -> QueryResult<()> {
    update(printing_task::table)
        .filter(printing_task::uuid.eq(uuid))
        .set((
            printing_task::done_instant.eq(new_done_instant),
        ))
        .execute(connection)?;
    Ok(())
}
