use chrono::prelude::*;
use diesel::insert_into;
use diesel::pg::PgConnection;
use diesel::prelude::*;
use serde::Serialize;

use crate::schema::{contest_problems, problem};

#[derive(Queryable)]
pub struct Problem {
    pub id: String,
    pub name: String,
    pub memory_limit_bytes: i32,
    pub time_limit_ms: i32,
    pub status: String,
    pub creation_user_id: i32,
    pub creation_instant: NaiveDateTime,
    pub repository_id: Option<i32>,
}

#[derive(Insertable)]
#[diesel(table_name = problem)]
pub struct NewProblem {
    pub id: String,
    pub name: String,
    pub memory_limit_bytes: i32,
    pub time_limit_ms: i32,
    pub status: String,
    pub creation_user_id: i32,
    pub creation_instant: NaiveDateTime,
}

#[derive(Queryable, Serialize)]
pub struct ProblemByContest {
    pub contest_id: i32,
    pub id: i32,
    pub name: String,
    pub label: String,
}

pub fn get_problems_by_contest_id(
    connection: &mut PgConnection,
    contest_id: i32,
) -> QueryResult<Vec<ProblemByContest>> {
    problem::table
        .inner_join(contest_problems::table)
        .filter(contest_problems::contest_id.eq(contest_id))
        .select((
            contest_problems::contest_id,
            contest_problems::id,
            problem::name,
            contest_problems::label,
        ))
        .order(contest_problems::label)
        .load(connection)
}

use diesel::sql_types;

#[derive(QueryableByName, Debug)]
pub struct ProblemByContestWithScore {
    #[diesel(sql_type = sql_types::Nullable<sql_types::Text>)]
    pub user_name: Option<String>,
    #[diesel(sql_type = sql_types::Nullable<sql_types::Text>)]
    pub user_full_name: Option<String>,
    #[diesel(sql_type = sql_types::Nullable<sql_types::Timestamp>)]
    pub first_ac_submission_instant: Option<NaiveDateTime>,
    #[diesel(sql_type = sql_types::Nullable<sql_types::Timestamp>)]
    pub first_ac_problem_submission_instant: Option<NaiveDateTime>,
    #[diesel(sql_type = sql_types::Integer)]
    pub failed_submissions: i32,
    #[diesel(sql_type = sql_types::Nullable<sql_types::Text>)]
    pub last_failed_submission_verdict: Option<String>,
    #[diesel(sql_type = sql_types::Integer)]
    pub id: i32,
    #[diesel(sql_type = sql_types::Text)]
    pub name: String,
    #[diesel(sql_type = sql_types::Text)]
    pub label: String,
    #[diesel(sql_type = sql_types::Integer)]
    pub memory_limit_bytes: i32,
    #[diesel(sql_type = sql_types::Integer)]
    pub time_limit_ms: i32,
    #[diesel(sql_type = sql_types::Integer)]
    pub user_accepted_count: i32,
    #[diesel(sql_type = sql_types::Nullable<sql_types::Text>)]
    pub balloon_color: Option<String>,
}

const WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT: &str = r#"
first_ac as (
    select
        min(submission_instant) as first_ac_submission_instant,
        contest_problem_id,
        filtered_submission.user_id
    from filtered_submission
    where filtered_submission.verdict = 'AC'
    group by filtered_submission.user_id, filtered_submission.contest_problem_id
), first_ac_by_problem as (
    select
        min(submission_instant) as first_ac_problem_submission_instant,
        contest_problem_id
    from filtered_submission
    where filtered_submission.verdict = 'AC'
    group by filtered_submission.contest_problem_id
), failed_submission_count as (
    select
        filtered_submission.user_id,
        filtered_submission.contest_problem_id,
        cast(count(*) as int) as count
    from filtered_submission
    left join first_ac on first_ac.contest_problem_id = filtered_submission.contest_problem_id
        and filtered_submission.user_id = first_ac.user_id
    where (
        first_ac_submission_instant is null or
        filtered_submission.submission_instant < first_ac.first_ac_submission_instant
    )
    group by filtered_submission.user_id, filtered_submission.contest_problem_id
), last_failed_submission_instant as (
    select
        filtered_submission.user_id,
        filtered_submission.contest_problem_id,
        max(filtered_submission.submission_instant) as last_submission_instant
    from filtered_submission
    left join first_ac on first_ac.contest_problem_id = filtered_submission.contest_problem_id
        and filtered_submission.user_id = first_ac.user_id
    where (
        first_ac_submission_instant is null or
        filtered_submission.submission_instant < first_ac.first_ac_submission_instant
    )
    group by filtered_submission.user_id, filtered_submission.contest_problem_id
), last_failed_submission as (
    select
        filtered_submission.user_id,
        filtered_submission.contest_problem_id,
        filtered_submission.verdict
    from filtered_submission
    inner join last_failed_submission_instant on last_failed_submission_instant.contest_problem_id = filtered_submission.contest_problem_id
        and last_failed_submission_instant.user_id = filtered_submission.user_id
        and last_failed_submission_instant.last_submission_instant = filtered_submission.submission_instant
), user_ac_count as (
    select
        cast(count(distinct filtered_submission.user_id) as int) as user_accepted_count,
        filtered_submission.contest_problem_id
    from filtered_submission
    where filtered_submission.verdict = 'AC'
    group by filtered_submission.contest_problem_id
)
"#;

const ALL_SUBMISSIONS: &str = r#"
filtered_submission as (select * from submission)
"#;

const ALL_SUBMISSIONS_IGNORING_CE_VERDICTS: &str = r#"
filtered_submission as (select * from submission where submission.verdict <> 'CE')
"#;

const BLIND_AND_FROZEN_SUBMISSIONS_IGNORING_CE_VERDICTS: &str = r#"
filtered_submission as (
    select
        submission.uuid,
        CASE WHEN submission.submission_instant < contest.blind_scoreboard_instant
            THEN submission.verdict ELSE 'BL' END as verdict,
        submission.submission_instant,
        submission.contest_problem_id,
        submission.user_id
    from submission
    inner join contest_problems on contest_problems.id = submission.contest_problem_id
    inner join contest on contest.id = contest_problems.contest_id
    where ((submission.user_id = $1) OR (submission.submission_instant < contest.frozen_scoreboard_instant))
    AND submission.verdict <> 'CE'
)
"#;

const FROZEN_SUBMISSIONS_IGNORING_CE_VERDICTS: &str = r#"
filtered_submission as (
    select
        submission.uuid,
        submission.verdict as verdict,
        submission.submission_instant,
        submission.contest_problem_id,
        submission.user_id
    from submission
    inner join contest_problems on contest_problems.id = submission.contest_problem_id
    inner join contest on contest.id = contest_problems.contest_id
    where submission.submission_instant < contest.frozen_scoreboard_instant
    and submission.verdict <> 'CE'
)
"#;

const SELECT_PROBLEMS_USER_WITH_SCORE: &str = r#"
    select
        "user".name as user_name,
        "user".full_name as user_full_name,
        first_ac_submission_instant,
        first_ac_problem_submission_instant,
        coalesce(failed_submission_count.count, 0) as failed_submissions,
        last_failed_submission.verdict as last_failed_submission_verdict,
        contest_problems.id,
        problem.name,
        contest_problems.label,
        problem.memory_limit_bytes,
        problem.time_limit_ms,
        coalesce(user_ac_count.user_accepted_count, 0) as user_accepted_count,
        contest_problems.balloon_color
    from contest_problems
    inner join problem on problem.id = contest_problems.problem_id
    inner join "user" on "user".id = $1
    left join last_failed_submission
        on last_failed_submission.contest_problem_id = contest_problems.id
        and last_failed_submission.user_id = "user".id
    left join failed_submission_count
        on failed_submission_count.contest_problem_id = contest_problems.id
        and failed_submission_count.user_id = "user".id
    left join first_ac
        on first_ac.contest_problem_id = contest_problems.id
        and first_ac.user_id = "user".id
    left join first_ac_by_problem
        on first_ac_by_problem.contest_problem_id = contest_problems.id
    left join user_ac_count
        on user_ac_count.contest_problem_id = contest_problems.id
"#;

// all verdicts for a given user
pub fn get_problems_user_with_score_without_blind_and_frozen(
    connection: &mut PgConnection,
    user_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
        order by problem.creation_instant
    "#,
        ALL_SUBMISSIONS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_USER_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(user_id)
    .load(connection)
}

pub fn get_problems_user_with_score_with_blind_and_frozen(
    connection: &mut PgConnection,
    user_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
        order by problem.creation_instant
    "#,
        BLIND_AND_FROZEN_SUBMISSIONS_IGNORING_CE_VERDICTS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_USER_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(user_id)
    .load(connection)
}

pub fn get_problems_user_by_contest_id_with_score_without_blind_and_frozen(
    connection: &mut PgConnection,
    user_id: i32,
    contest_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
        where contest_id = $2
        order by contest_problems.label
    "#,
        ALL_SUBMISSIONS_IGNORING_CE_VERDICTS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_USER_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(user_id)
    .bind::<sql_types::Integer, _>(contest_id)
    .load(connection)
}

pub fn get_problems_user_by_contest_id_with_score_with_blind_and_frozen(
    connection: &mut PgConnection,
    user_id: i32,
    contest_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
        where contest_id = $2
        order by contest_problems.label
    "#,
        BLIND_AND_FROZEN_SUBMISSIONS_IGNORING_CE_VERDICTS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_USER_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(user_id)
    .bind::<sql_types::Integer, _>(contest_id)
    .load(connection)
}

const SELECT_PROBLEMS_BY_CONTEST_ID_WITH_SCORE: &str = r#"
    (select
        null as user_name,
        null as user_full_name,
        null as first_ac_submission_instant,
        null as first_ac_problem_submission_instant,
        0 as failed_submissions,
        null as last_failed_submission_verdict,
        contest_problems.id,
        problem.name,
        contest_problems.label,
        problem.memory_limit_bytes,
        problem.time_limit_ms,
        0 as user_accepted_count,
        contest_problems.balloon_color
    from contest_problems
    inner join problem on problem.id = contest_problems.problem_id
    where contest_problems.contest_id = $2
    order by contest_problems.label)
    union all
    (select
        "user".name as user_name,
        "user".full_name as user_full_name,
        first_ac_submission_instant,
        first_ac_problem_submission_instant,
        coalesce(failed_submission_count.count, 0) as failed_submissions,
        last_failed_submission.verdict as last_failed_submission_verdict,
        contest_problems.id,
        problem.name,
        contest_problems.label,
        problem.memory_limit_bytes,
        problem.time_limit_ms,
        coalesce(user_ac_count.user_accepted_count, 0) as user_accepted_count,
        contest_problems.balloon_color
    from contest_problems
    inner join problem on problem.id = contest_problems.problem_id
    inner join (
        select user_id, contest_id
        from filtered_submission
        inner join contest_problems on contest_problems.id = filtered_submission.contest_problem_id
        group by user_id, contest_id
    ) as all_submitters on all_submitters.contest_id = $2
    inner join "user" on "user".id = all_submitters.user_id
    left join failed_submission_count
        on failed_submission_count.contest_problem_id = contest_problems.id
        and failed_submission_count.user_id = "user".id
    left join last_failed_submission
        on last_failed_submission.contest_problem_id = contest_problems.id
        and last_failed_submission.user_id = "user".id
    left join first_ac
        on first_ac.contest_problem_id = contest_problems.id
        and first_ac.user_id = "user".id
    left join first_ac_by_problem
        on first_ac_by_problem.contest_problem_id = contest_problems.id
    left join user_ac_count
        on user_ac_count.contest_problem_id = contest_problems.id
    where contest_problems.contest_id = $2
    order by "user".name, contest_problems.label)
"#;

pub fn get_problems_by_contest_id_with_score_without_blind_and_frozen(
    connection: &mut PgConnection,
    contest_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
    "#,
        ALL_SUBMISSIONS_IGNORING_CE_VERDICTS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_BY_CONTEST_ID_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(0)
    .bind::<sql_types::Integer, _>(contest_id)
    .load(connection)
}

pub fn get_problems_by_contest_id_with_score_with_blind_and_frozen(
    connection: &mut PgConnection,
    user_id: i32,
    contest_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
    "#,
        BLIND_AND_FROZEN_SUBMISSIONS_IGNORING_CE_VERDICTS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_BY_CONTEST_ID_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(user_id)
    .bind::<sql_types::Integer, _>(contest_id)
    .load(connection)
}

pub fn get_problems_by_contest_id_with_score_with_frozen(
    connection: &mut PgConnection,
    contest_id: i32,
) -> QueryResult<Vec<ProblemByContestWithScore>> {
    diesel::sql_query(format!(
        r#"
        with {}, {}
        {}
    "#,
        FROZEN_SUBMISSIONS_IGNORING_CE_VERDICTS,
        WITH_FIRST_AC_AND_FAILED_SUBMISSION_COUNT_AND_LAST_FAILED_SUBMISSION_AND_USER_AC_COUNT,
        SELECT_PROBLEMS_BY_CONTEST_ID_WITH_SCORE
    ))
    .bind::<sql_types::Integer, _>(0)
    .bind::<sql_types::Integer, _>(contest_id)
    .load(connection)
}

pub fn get_problem_by_contest_id_label(
    connection: &mut PgConnection,
    contest_id: i32,
    problem_label: &str,
) -> QueryResult<ProblemByContest> {
    problem::table
        .inner_join(contest_problems::table)
        .filter(contest_problems::contest_id.eq(contest_id))
        .filter(contest_problems::label.eq(problem_label))
        .select((
            contest_problems::contest_id,
            contest_problems::id,
            problem::name,
            contest_problems::label,
        ))
        .order(contest_problems::label)
        .first(connection)
}

#[derive(Queryable)]
pub struct ProblemByContestMetadata {
    pub contest_id: i32,
    pub label: String,
    pub id: String,
    pub memory_limit_bytes: i32,
    pub time_limit_ms: i32,
    pub status: String,
}

pub fn get_problem_by_contest_id_metadata(
    connection: &mut PgConnection,
    contest_problem_id: i32,
) -> QueryResult<ProblemByContestMetadata> {
    problem::table
        .inner_join(contest_problems::table)
        .filter(contest_problems::id.eq(contest_problem_id))
        .select((
            contest_problems::contest_id,
            contest_problems::label,
            problem::id,
            problem::memory_limit_bytes,
            problem::time_limit_ms,
            problem::status,
        ))
        .first(connection)
}

pub fn upsert_problem(
    connection: &mut PgConnection,
    new_problem: NewProblem,
) -> QueryResult<Problem> {
    match problem::table
        .filter(problem::id.eq(&new_problem.id))
        .first::<Problem>(connection)
        .optional()?
    {
        Some(p) => Ok(p),
        None => {
            insert_into(problem::table)
                .values(&new_problem)
                .execute(connection)?;
            problem::table
                .filter(problem::id.eq(&new_problem.id))
                .first::<Problem>(connection)
        }
    }
}
