use chrono::Local;
use std::io::{Cursor, Write};
use uuid::Uuid;

use actix_multipart::Multipart;
use futures::{StreamExt, TryStreamExt};

use crate::models::{backup, contest};
use crate::pages::assert_contest_not_started;
use crate::pages::prelude::*;

#[post("/contests/{id}/backups/")]
async fn create_backup(
    identity: Identity,
    mut payload: Multipart,
    pool: Data<DbPool>,
    path: Path<(i32,)>,
    request: HttpRequest,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let (contest_id,) = path.into_inner();
    let mut connection = pool.get()?;

    #[derive(Debug)]
    struct Form {
        contents: Option<Vec<u8>>,
        filename: Option<String>,
    }

    let mut form = Form {
        contents: None,
        filename: None,
    };

    while let Ok(Some(mut field)) = payload.try_next().await {
        let mut cursor = Cursor::new(vec![]);
        while let Some(chunk) = field.next().await {
            let data = chunk.unwrap();
            cursor
                .write(&data)
                .map_err(|_| PageError::Validation("Corpo inválido".into()))?;
        }

        cursor.set_position(0);

        let content_disposition = field.content_disposition();
        match content_disposition.get_name() {
            Some("contents") => {
                form.filename = Some(
                    content_disposition
                        .get_filename()
                        .ok_or(PageError::Validation(format!(
                            "Campo contents sem nome de arquivo"
                        )))?
                        .into(),
                );
                form.contents = Some(cursor.into_inner());
            }
            _ => {}
        }
    }

    if let None = form.contents {
        return Err(PageError::Validation("Arquivo não submetido".into()));
    }

    let contents = form.contents.unwrap();
    let filename = form.filename.unwrap();

    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    assert_contest_not_started(&logged_user, &contest)?;

    let uuid = Uuid::new_v4();
    backup::insert_backup(
        &mut connection,
        backup::NewBackup {
            uuid: uuid.to_string(),
            filename,
            contents,
            creation_instant: Local::now().naive_utc(),
            contest_id: contest.id,
            user_id: logged_user.id,
        },
    )?;

    Ok(redirect_to_referer(
        "Backup criado com sucesso".into(),
        &request,
    ))
}
