use chrono::Local;
use std::io::{Cursor, Read, Write};
use uuid::Uuid;

use actix_multipart::Multipart;
use futures::{StreamExt, TryStreamExt};

use crate::models::{contest, problem, submission};
use crate::pages::assert_contest_not_over;
use crate::pages::assert_contest_not_started;
use crate::pages::prelude::*;
use crate::queue::job_protocol::{job, Job, Language};

#[post("/contests/{id}/submissions/")]
async fn create_contest_submission(
    identity: Identity,
    mut payload: Multipart,
    pool: Data<DbPool>,
    job_sender: Data<Sender<Job>>,
    languages: Data<Arc<DashMap<String, Language>>>,
    session: Session,
    request: HttpRequest,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let mut connection = pool.get()?;

    #[derive(Debug)]
    struct Form {
        contest_problem_id: Option<i32>,
        language: Option<String>,
        source_text: Option<String>,
    }

    let mut form = Form {
        contest_problem_id: None,
        language: None,
        source_text: None,
    };

    while let Ok(Some(mut field)) = payload.try_next().await {
        let mut cursor = Cursor::new(vec![]);
        while let Some(chunk) = field.next().await {
            let data = chunk.unwrap();
            cursor
                .write(&data)
                .map_err(|_| PageError::Validation("Corpo inválido".into()))?;
        }

        cursor.set_position(0);

        fn parse_field(field: &str, cursor: &mut Cursor<Vec<u8>>) -> Result<String, PageError> {
            let mut value = String::new();
            cursor
                .read_to_string(&mut value)
                .map_err(|_| PageError::Validation(format!("Campo {} inválido", field)))?;
            Ok(value)
        }

        match field.content_disposition().get_name() {
            Some("language") => form.language = Some(parse_field("language", &mut cursor)?),
            Some("contest_problem_id") => {
                form.contest_problem_id =
                    parse_field("contest_problem_id", &mut cursor)?.parse().ok()
            }
            Some("source_text") => {
                form.source_text = Some(parse_field("source_text", &mut cursor)?)
            }
            _ => {}
        }
    }

    if let None = form.source_text {
        return Err(PageError::Validation("Código-fonte não submetido".into()));
    }

    let language = form.language.unwrap_or("".to_string());

    languages
        .get(&language)
        .ok_or_else(|| PageError::Validation("Linguagem inexistente".into()))?;

    let contest = contest::get_contest_by_contest_problem_id(
        &mut connection,
        form.contest_problem_id.unwrap(),
    )?;
    assert_contest_not_started(&logged_user, &contest)?;
    assert_contest_not_over(&logged_user, &contest)?;

    let source_text = form.source_text.unwrap();

    let uuid = Uuid::new_v4();
    submission::insert_submission(
        &mut connection,
        submission::NewSubmission {
            uuid: uuid.to_string(),
            source_text: source_text.clone(),
            language: language.clone(),
            submission_instant: Local::now().naive_utc(),
            contest_problem_id: form.contest_problem_id.unwrap(),
            user_id: logged_user.id,
        },
    )?;

    let metadata = problem::get_problem_by_contest_id_metadata(
        &mut connection,
        form.contest_problem_id.unwrap(),
    )?;

    job_sender
        .send(Job {
            uuid: uuid.to_string(),
            language: language.clone(),
            time_limit_ms: metadata.time_limit_ms,
            memory_limit_kib: metadata.memory_limit_bytes / 1_024,

            which: Some(job::Which::Judgement(job::Judgement {
                source_text: source_text.clone(),
                package_id: metadata.id,
            })),
        })
        .await?;

    session.insert("language", language.clone())?;
    Ok(redirect_to_referer(
        format!("Submetido {} com sucesso!", uuid),
        &request,
    ))
}
