use chrono::Local;
use std::io::{Cursor, Read, Write};
use uuid::Uuid;

use actix_multipart::Multipart;
use futures::{StreamExt, TryStreamExt};

use crate::models::{contest, printing_task};
use crate::pages::assert_contest_not_started;
use crate::pages::prelude::*;

#[post("/contests/{id}/printing-tasks/")]
async fn create_printing_task(
    identity: Identity,
    mut payload: Multipart,
    pool: Data<DbPool>,
    path: Path<(i32,)>,
    request: HttpRequest,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let (contest_id,) = path.into_inner();
    let mut connection = pool.get()?;

    #[derive(Debug)]
    struct Form {
        contents: Option<String>,
    }

    let mut form = Form { contents: None };

    while let Ok(Some(mut field)) = payload.try_next().await {
        let mut cursor = Cursor::new(vec![]);
        while let Some(chunk) = field.next().await {
            let data = chunk.unwrap();
            cursor
                .write(&data)
                .map_err(|_| PageError::Validation("Corpo inválido".into()))?;
        }

        cursor.set_position(0);

        fn parse_field(field: &str, cursor: &mut Cursor<Vec<u8>>) -> Result<String, PageError> {
            let mut value = String::new();
            cursor
                .read_to_string(&mut value)
                .map_err(|_| PageError::Validation(format!("Campo {} inválido", field)))?;
            Ok(value)
        }

        match field.content_disposition().get_name() {
            Some("contents") => form.contents = Some(parse_field("contents", &mut cursor)?),
            _ => {}
        }
    }

    if let None = form.contents {
        return Err(PageError::Validation("Arquivo não submetido".into()));
    }

    let contents = form.contents.unwrap();

    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    assert_contest_not_started(&logged_user, &contest)?;

    let uuid = Uuid::new_v4();
    printing_task::insert_printing_task(
        &mut connection,
        printing_task::NewPrintingTask {
            uuid: uuid.to_string(),
            contents: contents.clone(),
            creation_instant: Local::now().naive_utc(),
            status: "pending".into(),
            done_instant: None,
            contest_id: contest.id,
            user_id: logged_user.id,
        },
    )?;

    Ok(redirect_to_referer(
        "Tarefa de impressão criada com sucesso".into(),
        &request,
    ))
}
