use crate::models::{contest, clarification};
use crate::pages::prelude::*;
use crate::pages::{
    assert_contest_not_started, assert_not_single_contest, get_formatted_contest,
    FormattedContest,
};

#[derive(Serialize)]
struct FormattedClarification {
    pub uuid: String,
    pub question: String,
    pub question_instant: String,
    pub answer: Option<String>,
    pub answer_instant: Option<String>,
    pub answer_user_id: Option<i32>,
    pub answer_to_all: bool,
    pub contest_problem_id: Option<i32>,
    pub problem_label: Option<String>,
    pub contest_id: i32,
    pub user_id: i32,
    pub user_name: String,
    pub user_full_name: Option<String>,
}

#[get("/contests/{id}/clarifications")]
pub async fn get_contest_clarifications_by_id(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    tz: Data<Tz>,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let (contest_id,) = path.into_inner();

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        contest: FormattedContest,
        clarifications: Vec<FormattedClarification>,
    }

    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    assert_contest_not_started(&logged_user, &contest)?;
    if let Some(redirect) = assert_not_single_contest(&logged_user) {
        return redirect;
    }

    let clarifications = if !logged_user.is_admin {
        clarification::get_clarifications_by_contest_user(
            &mut connection,
            logged_user.id,
            contest_id,
        )?
    } else {
        clarification::get_clarifications_by_contest(
            &mut connection,
            contest_id,
        )?
    };

    render(
        &hb,
        "clarifications",
        &Context {
            base,
            contest: get_formatted_contest(&tz, &contest),
            clarifications: clarifications
                .into_iter()
                .map(|(c, u, problem_label)| FormattedClarification {
                    uuid: c.uuid,
                    question: c.question,
                    question_instant: format_utc_date_time(&tz, c.question_instant),
                    answer: c.answer,
                    answer_instant: c.answer_instant.map(|i| format_utc_date_time(&tz, i)),
                    answer_user_id: c.answer_user_id,
                    answer_to_all: c.answer_to_all,
                    contest_problem_id: c.contest_problem_id,
                    problem_label,
                    contest_id: c.contest_id,
                    user_id: c.user_id,
                    user_name: u.name,
                    user_full_name: u.full_name,
                })
                .collect(),
        },
    )
}
