use itertools::Itertools;

use crate::models::{contest, problem, submission};
use crate::pages::prelude::*;
use crate::pages::{
    assert_contest_not_started, assert_not_single_contest, get_formatted_contest,
    get_formatted_problem_by_contest_with_score, get_formatted_submissions,
    get_formatted_submissions_with_blind, FormattedContest, FormattedProblemByContestWithScore,
    FormattedSubmission,
};

#[get("/contests/{id}/scoreboard")]
async fn get_contest_scoreboard_by_id(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    tz: Data<Tz>,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    if let Some(redirect) = assert_not_single_contest(&logged_user) {
        return redirect;
    }
    let (contest_id,) = path.into_inner();

    #[derive(Serialize)]
    struct Score {
        pub user_name: Option<String>,
        pub user_full_name: Option<String>,
        pub problems: Vec<FormattedProblemByContestWithScore>,
        pub solved_count: i64,
        pub penalty: i64,
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        contest: FormattedContest,
        scores: Vec<Score>,
        submissions: Vec<FormattedSubmission>,
    }

    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    assert_contest_not_started(&logged_user, &contest)?;
    let scores = if logged_user.is_admin {
        problem::get_problems_by_contest_id_with_score_without_blind_and_frozen(&mut connection, contest_id)?
    } else {
        problem::get_problems_by_contest_id_with_score_with_frozen(&mut connection, contest_id)?
    };
    let submissions =
        submission::get_submissions_user_by_contest(&mut connection, logged_user.id, contest_id)?;
    let mut scores: Vec<_> = scores.into_iter().group_by(|e| (e.user_name.as_ref().cloned(), e.user_full_name.as_ref().cloned())).into_iter().map(|((user_name, user_full_name), problems)| {
        let problems: Vec<_> = problems.map(|p| get_formatted_problem_by_contest_with_score(&p, &contest)).collect();

        Score {
            user_name,
            user_full_name,
            solved_count: i64::try_from(problems.iter().filter(|p| !p.first_ac_submission_time.is_empty()).count()).unwrap(),
            penalty: problems.iter().filter(|p| !p.first_ac_submission_time.is_empty())
                .map(|p| match p.first_ac_submission_minutes {
                    Some(x) if x >= 0 => x,
                    _ => 0,
                } + i64::from(20*p.failed_submissions))
                .sum(),
                problems
        }
    }).collect();
    scores.sort_by(|a, b| {
        (
            a.user_name.is_some(),
            -a.solved_count,
            a.penalty,
            &a.user_name,
        )
            .cmp(&(
                b.user_name.is_some(),
                -b.solved_count,
                b.penalty,
                &b.user_name,
            ))
    });

    render(
        &hb,
        "scoreboard",
        &Context {
            base,
            contest: get_formatted_contest(&tz, &contest),
            scores,
            submissions: if !logged_user.is_admin {
                get_formatted_submissions_with_blind(&tz, &contest, &submissions)
            } else {
                get_formatted_submissions(&tz, &submissions)
            },
        },
    )
}
