use chrono::NaiveDateTime;
use std::collections::HashSet;

use crate::models::{contest, clarification, submission, printing_task};
use crate::pages::prelude::*;
use crate::pages::{get_formatted_contest, FormattedContest,
    get_formatted_submissions, FormattedSubmission,
};

#[get("/contests/{id}/tasks")]
pub async fn get_contest_tasks_by_id(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    tz: Data<Tz>,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    if !logged_user.is_admin {
        return Err(PageError::Forbidden(
            "Apenas administradores podem fazer isso".into(),
        ));
    }
    let (contest_id,) = path.into_inner();

    struct Task {
        uuid: String,
        creation_instant: NaiveDateTime,
        t: String,
        user_name: String,
        user_full_name: Option<String>,
        description: String,
        balloon_color: Option<String>,
        first_solve: bool,
        done: bool
    }

    #[derive(Serialize)]
    struct FormattedTask {
        uuid: String,
        creation_instant: String,
        t: String,
        user_name: String,
        user_full_name: Option<String>,
        description: String,
        balloon_color: Option<String>,
        first_solve: bool,
        done: bool
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        contest: FormattedContest,
        tasks: Vec<FormattedTask>,
        submissions: Vec<FormattedSubmission>,
    }

    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;

    let mut tasks: Vec<Task> = vec![];

    let clarifications = clarification::get_clarifications_by_contest(&mut connection, contest_id)?;
    for (clarification, user, problem_label) in clarifications {
        tasks.push(Task {
            uuid: clarification.uuid,
            creation_instant: clarification.question_instant,
            t: "clarification".into(),
            user_name: user.name,
            user_full_name: user.full_name,
            balloon_color: None,
            description: match problem_label {
                Some(problem_label) => format!("Problema {}", problem_label),
                None => "Geral".into()
            },
            first_solve: false,
            done: clarification.answer_instant.is_some()
        });
    }

    let mut submissions = submission::get_submissions_by_contest(&mut connection, contest_id)?;
    submissions.reverse();
    let mut solved_contest_problems: HashSet<i32> = Default::default();
    let mut solved_contest_problems_by_user: HashSet<(i32, i32)> = Default::default();
    for (submission, contest_problem, user) in &submissions {
        if submission.verdict.as_ref().map(|e| e != "AC").unwrap_or(true) { continue }
        if solved_contest_problems_by_user.contains(&(contest_problem.id, user.id)) {
            continue
        }
        // if contest.frozen_scoreboard_instant.map(|e| submission.submission_instant >= e).unwrap_or(false) {
        //     continue
        // }
        tasks.push(Task {
            uuid: submission.uuid.clone(),
            creation_instant: submission.submission_instant,
            t: "balloon".into(),
            user_name: user.name.clone(),
            user_full_name: user.full_name.clone(),
            balloon_color: contest_problem.balloon_color.clone(),
            description: format!("Problema {}", contest_problem.label),
            first_solve: !solved_contest_problems.contains(&contest_problem.id),
            done: submission.balloon_delivered_instant.is_some(),
        });
        solved_contest_problems.insert(contest_problem.id);
        solved_contest_problems_by_user.insert((contest_problem.id, user.id));
    }

    let printing_tasks = printing_task::get_printing_tasks_by_contest(
        &mut connection,
        contest_id
    )?;
    for (printing_task, user) in printing_tasks {
        tasks.push(Task {
            uuid: printing_task.uuid.clone(),
            creation_instant: printing_task.creation_instant,
            t: "printing".into(),
            user_name: user.name.clone(),
            user_full_name: user.full_name.clone(),
            balloon_color: None,
            description: format!("{} bytes", printing_task.contents.len()),
            first_solve: false,
            done: printing_task.done_instant.is_some(),
        });
    }

    tasks.sort_by(|a, b| {
        (a.done, b.creation_instant).cmp(&(b.done, a.creation_instant))
    });

    render(
        &hb,
        "tasks",
        &Context {
            base,
            contest: get_formatted_contest(&tz, &contest),
            tasks: tasks.into_iter().map(|e| FormattedTask {
                uuid: e.uuid,
                creation_instant: format_utc_date_time(&tz, e.creation_instant),
                t: e.t,
                user_name: e.user_name,
                user_full_name: e.user_full_name,
                balloon_color: e.balloon_color,
                description: e.description,
                first_solve: e.first_solve,
                done: e.done
            }).collect(),
            submissions: get_formatted_submissions(&tz, &submissions),
        },
    )
}
