use std::collections::HashSet;
use zip::ZipWriter;
use zip::write::FileOptions;
use std::io::Write;
use std::cmp::min;
use chrono::Local;
use regex::Regex;

use crate::models::{submission, contest, problem, user};
use crate::pages::prelude::*;

#[derive(Deserialize)]
struct SecretParams {
    secret: String,
}

#[get("/contests/{id}/webcast.zip")]
async fn get_contest_webcast_by_id(
    pool: Data<DbPool>,
    path: Path<(i32,)>,
    query: Query<SecretParams>
) -> PageResult {
    let (contest_id,) = path.into_inner();

    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    let problems = problem::get_problems_by_contest_id(&mut connection, contest_id)?;
    let users_contest = user::get_users_by_single_contest_id(&mut connection, contest_id)?;
    let users_contest_users: HashSet<String> =
        HashSet::from_iter(users_contest.iter().map(|e| e.name.clone()));
    let submissions = submission::get_submissions_by_contest(&mut connection, contest_id)?;

    let mut buf = [0; 65536];
    {
        let mut zip = ZipWriter::new(std::io::Cursor::new(&mut buf[..]));

        let options = FileOptions::default().compression_method(zip::CompressionMethod::Stored);

        zip.start_file("contest", options)?;
        zip.write(contest.name.as_bytes())?;
        zip.write(b"\n")?;
        let start_instant = contest.start_instant.unwrap();
        let end_instant = contest.end_instant.unwrap();
        let maximum_time = (end_instant - start_instant).num_minutes();
        let current_time = min(maximum_time, (Local::now().naive_utc() - start_instant).num_minutes());
        let score_freeze_time = match contest.frozen_scoreboard_instant {
            None => maximum_time,
            Some(frozen_scoreboard_instant) => (frozen_scoreboard_instant - start_instant).num_minutes()
        };
        let penalty = 20;
        let contest_params = format!("{}\x1c{}\x1c{}\x1c{}\n", maximum_time, current_time, score_freeze_time, penalty);
        zip.write(contest_params.as_bytes())?;
        let number_teams = users_contest.len();
        let number_problems = problems.len();
        let team_params = format!("{}\x1c{}\n", number_teams, number_problems);
        zip.write(team_params.as_bytes())?;
        let re = Regex::new(r"^\[([^\]]+)\]").unwrap();
        for user in users_contest {
            let institution = user.full_name.as_ref().and_then(|e| re.captures(e))
                .and_then(|e| e.get(1)).map(|e| e.as_str()).unwrap_or("".into());
            let user_full_name = user.full_name.clone().unwrap_or("".into());
            zip.write(format!("{}\x1c{}\x1c{}\n", user.name, institution, user_full_name).as_bytes())?;
        }
        zip.write(b"1\x1c1\n")?;
        zip.write(format!("{}\x1cY\n", number_problems).as_bytes())?;

        let mut i = 1;
        zip.start_file("runs", options)?;
        for (submission, contest_problem, user) in submissions {
            if !users_contest_users.contains(&user.name) { continue }
            zip.write(format!(
                "{}\x1c{}\x1c{}\x1c{}\x1c{}\n",
                (submission.submission_instant.and_utc().timestamp())*10 + i,
                (submission.submission_instant - start_instant).num_minutes(),
                user.name,
                contest_problem.label,
                match submission.verdict.as_ref().map(|e| e.as_str()) {
                    Some("WJ") | None => "?",
                    Some("AC") => "Y",
                    Some("CE") => "X",
                    _ => "N",
                }
            ).as_bytes())?;
            i += 1;
            if i > 10 { i -= 10; }
        }

        zip.start_file("time", options)?;
        zip.write(format!("{}", min((end_instant - start_instant).num_seconds(), (Local::now().naive_utc() - start_instant).num_seconds())).as_bytes())?;

        zip.start_file("version", options)?;
        zip.write(b"1.0")?;

        zip.finish()?;
    }

    Ok(HttpResponse::Ok().body(Vec::from(buf)))
}
