use crate::models::repository;
use crate::pages::prelude::*;
use std::fs;
use toml;
use walkdir::WalkDir;

#[derive(Deserialize)]
struct BlemConfig {
    authors: Vec<String>,
    time_limit_ms: i32,
    memory_limit_kb: Option<i32>,
    br: LanguageConfig,
}

#[derive(Deserialize)]
struct LanguageConfig {
    name: String,
}

#[derive(Deserialize)]
struct ContestProblemConfig {
    label: String,
    path: String,
    name: String,
}

#[derive(Deserialize)]
struct ContestConfig {
    languages: Vec<String>,
    br: LanguageConfig,
    problems: Vec<ContestProblemConfig>,
}

#[get("/repositories/{id}")]
pub async fn get_repository_by_id(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    tz: Data<Tz>,
) -> PageResult {
    use std::path::Path;

    let logged_user = require_identity(&identity)?;
    if !logged_user.is_admin {
        return Err(PageError::Forbidden(
            "Apenas administradores podem fazer isso".into(),
        ));
    }
    let (repository_id,) = path.into_inner();

    #[derive(Serialize)]
    struct FormattedRepository {
        pub id: i32,
        pub name: String,
        pub remote_url: String,
        pub creation_instant: String,
    }

    #[derive(Serialize)]
    struct RepositoryProblem {
        pub name: String,
        pub authors: Vec<String>,
        pub time_limit_ms: i32,
        pub memory_limit_kb: i32,
        pub path: String,
    }

    #[derive(Serialize)]
    struct RepositoryContestProblem {
        pub label: String,
        pub name: String,
        pub path: String,
        pub time_limit_ms: Option<i32>,
        pub authors: Vec<String>,
        pub last_commit: String,
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        repository: FormattedRepository,
        problems: Vec<RepositoryProblem>,
        contest_name: Option<String>,
        contest_problems: Vec<RepositoryContestProblem>,
    }

    let mut connection = pool.get()?;
    let repository = repository::get_repository_by_id(&mut connection, repository_id)?;
    let git_repository =
        git2::Repository::open(Path::new(&format!("/data/{}/", repository.path))).unwrap();
    let head_commit = git_repository.head().unwrap();
    let head_commit_short_id = head_commit
        .peel_to_commit()
        .unwrap()
        .as_object()
        .short_id()
        .unwrap();
    let head_commit_shorthand = head_commit_short_id.as_str().unwrap();

    let mut contest_name = None;
    let mut contest_problems = Vec::new();
    let maybe_contest_toml_contents =
        fs::read_to_string(format!("/data/{}/contest.toml", repository.path));
    if let Ok(contest_toml_contents) = maybe_contest_toml_contents {
        let maybe_contest_config = toml::from_str::<ContestConfig>(&contest_toml_contents);
        if let Ok(contest_config) = maybe_contest_config {
            contest_name = Some(contest_config.br.name);
            contest_problems = contest_config
                .problems
                .into_iter()
                .map(|c| RepositoryContestProblem {
                    label: c.label,
                    name: c.name,
                    path: c.path,
                    time_limit_ms: None,
                    last_commit: head_commit_shorthand.into(),
                    authors: vec![],
                })
                .collect();
        }
    }

    let mut problems = Vec::new();
    for entry in WalkDir::new(format!("/data/{}", repository.path))
        .into_iter()
        .filter_entry(|e| e.file_name().to_str().map(|s| s != "blem").unwrap_or(false))
        .filter_map(|e| e.ok().filter(|e| e.file_name() == "blem.toml"))
    {
        let maybe_blem_config = toml::from_str::<BlemConfig>(&fs::read_to_string(entry.path())?);
        let path = entry
            .path()
            .strip_prefix(format!("/data/{}", repository.path))
            .unwrap()
            .parent()
            .unwrap()
            .to_str()
            .unwrap()
            .into();
        if let Ok(blem_config) = maybe_blem_config {
            let mut found = false;
            for contest_problem in &mut contest_problems {
                if contest_problem.path == path {
                    contest_problem.name =
                        format!("{0} ({1})", contest_problem.name, blem_config.br.name);
                    contest_problem.time_limit_ms = Some(blem_config.time_limit_ms);
                    contest_problem.authors = blem_config.authors.clone();
                    found = true;
                }
            }
            if !found {
                problems.push(RepositoryProblem {
                    name: blem_config.br.name,
                    time_limit_ms: blem_config.time_limit_ms,
                    memory_limit_kb: blem_config.memory_limit_kb.unwrap_or(204800),
                    authors: blem_config.authors,
                    path,
                });
            }
        } else {
            problems.push(RepositoryProblem {
                name: "???".to_string(),
                time_limit_ms: 0,
                memory_limit_kb: 0,
                authors: vec![],
                path,
            });
        }
    }

    render(
        &hb,
        "repository",
        &Context {
            base,
            problems,
            contest_name,
            contest_problems,
            repository: FormattedRepository {
                id: repository.id,
                name: repository.name.clone(),
                remote_url: repository.remote_url.clone(),
                creation_instant: format_utc_date_time(&tz, repository.creation_instant),
            },
        },
    )
}
