use crate::models::repository;
use crate::pages::prelude::*;
use std::fs;
use toml;

#[derive(Deserialize)]
struct BlemConfig {
    authors: Vec<String>,
    time_limit_ms: i32,
    br: LanguageConfig,
}

#[derive(Deserialize)]
struct LanguageConfig {
    name: String,
}

#[get("/repositories/{id}/problems-by-path/{path:.*}")]
pub async fn get_repository_problem_by_path(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32, String)>,
    tz: Data<Tz>,
) -> PageResult {
    use std::path::Path;

    let logged_user = require_identity(&identity)?;
    if !logged_user.is_admin {
        return Err(PageError::Forbidden(
            "Apenas administradores podem fazer isso".into(),
        ));
    }
    let (repository_id, problem_path) = path.into_inner();

    #[derive(Serialize)]
    struct FormattedRepository {
        pub id: i32,
        pub name: String,
        pub remote_url: String,
        pub creation_instant: String,
    }

    #[derive(Serialize)]
    struct RepositoryProblem {
        pub name: String,
        pub authors: Vec<String>,
        pub time_limit_ms: i32,
        pub path: String,
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        repository: FormattedRepository,
        problems: Vec<RepositoryProblem>,
    }

    let mut connection = pool.get()?;
    let repository = repository::get_repository_by_id(&mut connection, repository_id)?;

    let path = Path::new(&format!("/data/{}", repository.path))
        .join(problem_path)
        .canonicalize()
        .unwrap();
    if !path.starts_with(format!("/data/{}", repository.path)) {
        return Err(PageError::Forbidden(
            "Caminho para o problema deve estar dentro do repositório".into(),
        ));
    }

    let maybe_blem_config =
        toml::from_str::<BlemConfig>(&fs::read_to_string(path.join("blem.toml"))?);
    let mut problems = Vec::new();
    if let Ok(blem_config) = maybe_blem_config {
        problems.push(RepositoryProblem {
            name: blem_config.br.name,
            time_limit_ms: blem_config.time_limit_ms,
            authors: blem_config.authors,
            path: path.to_str().unwrap().into(),
        });
    }

    render(
        &hb,
        "repository",
        &Context {
            base,
            problems,
            repository: FormattedRepository {
                id: repository.id,
                name: repository.name.clone(),
                remote_url: repository.remote_url.clone(),
                creation_instant: format_utc_date_time(&tz, repository.creation_instant),
            },
        },
    )
}
