use crate::models::{clarification, contest, problem, submission};
use crate::pages::prelude::*;
use crate::pages::{
    assert_single_contest_not_started, get_formatted_contest,
    get_formatted_submissions_with_blind_ascending, FormattedContest, FormattedSubmission,
};

#[derive(Serialize)]
struct FormattedClarification {
    pub uuid: String,
    pub question: String,
    pub question_instant: String,
    pub answer: Option<String>,
    pub answer_instant: Option<String>,
    pub answer_user_id: Option<i32>,
    pub answer_to_all: bool,
    pub contest_problem_id: Option<i32>,
    pub problem_label: Option<String>,
    pub contest_id: i32,
    pub user_id: i32,
    pub user_name: String,
}

#[get("/single-contest/{id}/clarifications")]
async fn get_single_contest_clarifications(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    tz: Data<Tz>,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let (contest_id,) = path.into_inner();

    #[derive(Serialize)]
    struct FormattedProblem {
        pub id: i32,
        pub label: String,
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        contest: FormattedContest,
        problems: Vec<FormattedProblem>,
        clarifications: Vec<FormattedClarification>,
        submissions: Vec<FormattedSubmission>,
    }

    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    if let Some(redirect) = assert_single_contest_not_started(&logged_user, &contest) {
        return redirect;
    }
    let submissions =
        submission::get_submissions_user_by_contest(&mut connection, logged_user.id, contest_id)?;
    let clarifications = clarification::get_clarifications_by_contest_user(
        &mut connection,
        logged_user.id,
        contest_id,
    )?;
    let problems = problem::get_problems_by_contest_id(&mut connection, contest_id)?;

    render(
        &hb,
        "single_clarifications",
        &Context {
            base,
            problems: problems
                .iter()
                .map(|p| FormattedProblem {
                    id: p.id,
                    label: p.label.clone(),
                })
                .collect(),
            contest: get_formatted_contest(&tz, &contest),
            clarifications: clarifications
                .into_iter()
                .map(|(c, u, problem_label)| FormattedClarification {
                    uuid: c.uuid,
                    question: c.question,
                    question_instant: format_utc_date_time(&tz, c.question_instant),
                    answer: c.answer,
                    answer_instant: c.answer_instant.map(|i| format_utc_date_time(&tz, i)),
                    answer_user_id: c.answer_user_id,
                    answer_to_all: c.answer_to_all,
                    contest_problem_id: c.contest_problem_id,
                    problem_label,
                    contest_id: c.contest_id,
                    user_id: c.user_id,
                    user_name: u.name,
                })
                .collect(),
            submissions: get_formatted_submissions_with_blind_ascending(
                &tz,
                &contest,
                &submissions,
            ),
        },
    )
}
