use crate::models::contest;
use crate::models::problem;
use crate::models::submission;
use crate::pages::get_editor::{get_formatted_languages, FormattedLanguage};
use crate::pages::prelude::*;
use crate::pages::{
    assert_single_contest_not_started, get_formatted_contest,
    get_formatted_submissions_with_blind_ascending, FormattedContest, FormattedSubmission,
};
use crate::Language;

#[derive(Deserialize)]
struct ContestProblemHint {
    contest_problem_id: Option<i32>,
}

#[get("/single-contest/{id}/runs")]
async fn get_single_contest_runs(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    languages: Data<Arc<DashMap<String, Language>>>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    contest_problem_hint: Query<ContestProblemHint>,
    session: Session,
    tz: Data<Tz>,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let (contest_id,) = path.into_inner();
    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    if let Some(redirect) = assert_single_contest_not_started(&logged_user, &contest) {
        return redirect;
    }

    let submissions =
        submission::get_submissions_user_by_contest(&mut connection, logged_user.id, contest_id)?;

    let problems = problem::get_problems_by_contest_id(&mut connection, contest_id)?;

    #[derive(Serialize)]
    struct FormattedProblem {
        pub id: i32,
        pub label: String,
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        contest: FormattedContest,
        problems: Vec<FormattedProblem>,
        submissions: Vec<FormattedSubmission>,
        language: Option<String>,
        languages: Vec<FormattedLanguage>,
        contest_problem_id_hint: Option<i32>,
    }

    render(
        &hb,
        "single_runs",
        &Context {
            base,
            contest: get_formatted_contest(&tz, &contest),
            problems: problems
                .iter()
                .map(|p| FormattedProblem {
                    id: p.id,
                    label: p.label.clone(),
                })
                .collect(),
            submissions: get_formatted_submissions_with_blind_ascending(
                &tz,
                &contest,
                &submissions,
            ),
            language: session.get("language")?,
            languages: get_formatted_languages(&languages, true),
            contest_problem_id_hint: contest_problem_hint.contest_problem_id,
        },
    )
}
