use itertools::Itertools;
use std::collections::HashSet;
use chrono::Local;

use crate::models::{contest, problem, submission, user};
use crate::pages::prelude::*;
use crate::pages::{
    assert_single_contest_not_started, get_formatted_contest,
    get_formatted_problem_by_contest_with_score, get_formatted_submissions_with_blind_ascending,
    FormattedContest, FormattedProblemByContestWithScore, FormattedSubmission,
};

#[get("/single-contest/{id}/score")]
async fn get_single_contest_score(
    base: BaseContext,
    identity: Identity,
    pool: Data<DbPool>,
    hb: Data<Handlebars<'_>>,
    path: Path<(i32,)>,
    tz: Data<Tz>,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    let (contest_id,) = path.into_inner();

    #[derive(Serialize)]
    struct Score {
        pub user_name: Option<String>,
        pub user_full_name: Option<String>,
        pub problems: Vec<FormattedProblemByContestWithScore>,
        pub solved_count: i64,
        pub penalty: i64,
    }

    #[derive(Serialize)]
    struct Context {
        base: BaseContext,
        contest: FormattedContest,
        scores: Vec<Score>,
        submissions: Vec<FormattedSubmission>,
        frozen: bool,
    }

    let mut connection = pool.get()?;
    let contest = contest::get_contest_by_id(&mut connection, contest_id)?;
    if let Some(redirect) = assert_single_contest_not_started(&logged_user, &contest) {
        return redirect;
    }
    let scores =
        problem::get_problems_by_contest_id_with_score_with_frozen(&mut connection, contest_id)?;
    let submissions =
        submission::get_submissions_user_by_contest(&mut connection, logged_user.id, contest_id)?;
    let users_contest = user::get_users_by_single_contest_id(&mut connection, contest_id)?;
    let users_contest_users: HashSet<String> =
        HashSet::from_iter(users_contest.iter().map(|e| e.name.clone()));
    let mut scores: Vec<_> = scores.into_iter().group_by(|e|
        (e.user_name.as_ref().cloned(), e.user_full_name.clone())
    ).into_iter().filter_map(|((user_name, user_full_name), problems)| {
        let problems: Vec<_> = problems.map(|p| get_formatted_problem_by_contest_with_score(&p, &contest)).collect();
        if match user_name {
            None => false,
            Some(ref user_name) => !users_contest_users.contains(user_name)
        } {
            return None
        }
        Some(Score {
            user_name,
            user_full_name,
            solved_count: i64::try_from(problems.iter().filter(|p| !p.first_ac_submission_time.is_empty()).count()).unwrap(),
            penalty: problems.iter().filter(|p| !p.first_ac_submission_time.is_empty())
                .map(|p| match p.first_ac_submission_minutes {
                    Some(x) if x >= 0 => x,
                    _ => 0,
                } + i64::from(20*p.failed_submissions))
                .sum(),
            problems
        })
    }).collect();
    scores.sort_by(|a, b| {
        (
            a.user_name.is_some(),
            -a.solved_count,
            a.penalty,
            &a.user_name,
        )
            .cmp(&(
                b.user_name.is_some(),
                -b.solved_count,
                b.penalty,
                &b.user_name,
            ))
    });

    let scored_users: HashSet<String> =
        HashSet::from_iter(scores.iter().filter_map(|e| e.user_name.clone()));
    for user in users_contest {
        if !scored_users.contains(&user.name) {
            scores.push(Score {
                user_name: Some(user.name),
                user_full_name: user.full_name,
                solved_count: 0,
                penalty: 0,
                problems: scores[0]
                    .problems
                    .iter()
                    .map(|e| FormattedProblemByContestWithScore {
                        first_ac_submission_time: e.first_ac_submission_time.clone(),
                        first_ac_submission_minutes: e.first_ac_submission_minutes,
                        first_solve: false,
                        user_accepted_count: e.user_accepted_count,
                        failed_submissions: e.failed_submissions,
                        submission_count: e.submission_count,
                        last_failed_submission_verdict: None,
                        balloon_color: e.balloon_color.clone(),
                        id: e.id,
                        name: e.name.clone(),
                        label: e.label.clone(),
                        memory_limit_mib: e.memory_limit_mib,
                        time_limit: e.time_limit.clone(),
                    })
                    .collect(),
            });
        }
    }

    render(
        &hb,
        "single_score",
        &Context {
            base,
            contest: get_formatted_contest(&tz, &contest),
            scores,
            submissions: get_formatted_submissions_with_blind_ascending(
                &tz,
                &contest,
                &submissions,
            ),
            frozen: match contest.frozen_scoreboard_instant {
                Some(frozen_scoreboard_instant) => Local::now().naive_utc() >= frozen_scoreboard_instant,
                None => false
            }
        },
    )
}
