use crate::models::repository;
use crate::pages::prelude::*;
use chrono::Local;
use git2::{CertificateCheckStatus, Cred, RemoteCallbacks};
use itertools::Itertools;
use std::path::Path;

#[derive(Serialize, Deserialize)]
struct RepositoryForm {
    name: String,
    remote_url: String,
}

#[post("/repositories/")]
async fn import_repository(
    identity: Identity,
    pool: Data<DbPool>,
    form: Form<RepositoryForm>,
    request: HttpRequest,
) -> PageResult {
    let logged_user = require_identity(&identity)?;
    if !logged_user.is_admin {
        return Err(PageError::Forbidden(
            "Apenas administradores podem fazer isso".into(),
        ));
    }
    let mut connection = pool.get()?;

    let (_, url) = form.remote_url.split("@").collect_tuple().unwrap();
    let path = url.replace(':', "+").replace('/', ".");

    let mut callbacks = RemoteCallbacks::new();
    callbacks.credentials(|_url, username_from_url, _allowed_types| {
        Cred::ssh_key(
            username_from_url.unwrap(),
            None,
            Path::new("/data/id_ed25519_git"),
            None,
        )
    });
    callbacks.certificate_check(|_cert, _hostname| {
        return Ok(CertificateCheckStatus::CertificateOk);
    });

    let mut fo = git2::FetchOptions::new();
    fo.remote_callbacks(callbacks);

    let mut builder = git2::build::RepoBuilder::new();
    builder.fetch_options(fo);

    if let Err(e) = builder.clone(&form.remote_url, Path::new(&format!("/data/{}/", path))) {
        return Err(PageError::Validation(format!(
            "Repositório {} importado com erro {}",
            form.remote_url, e
        )));
    }

    repository::insert_repository(
        &mut connection,
        repository::NewRepository {
            name: form.name.clone(),
            path,
            remote_url: form.remote_url.clone(),
            creation_user_id: logged_user.id,
            creation_instant: Local::now().naive_utc(),
        },
    )?;

    Ok(redirect_to_referer(
        format!("Repositório {} importado com sucesso!", form.remote_url),
        &request,
    ))
}
