use crate::models::contest;
use crate::models::contest::NewContest;
use crate::models::contest::NewContestProblems;
use crate::models::problem;
use crate::models::problem::NewProblem;
use crate::models::repository;
use crate::pages::prelude::*;
use chrono::Local;
use diesel::NotFound;
use std::fs;
use std::process::Command;

#[derive(Deserialize)]
struct BlemConfig {
    authors: Vec<String>,
    time_limit_ms: i32,
    memory_limit_kb: Option<i32>,
    br: LanguageConfig,
}

#[derive(Deserialize)]
struct LanguageConfig {
    name: String,
}

#[derive(Deserialize)]
struct ContestProblemConfig {
    label: String,
    path: String,
    name: String,
}

#[derive(Deserialize)]
struct ContestConfig {
    br: LanguageConfig,
    problems: Vec<ContestProblemConfig>,
}

#[post("/repositories/{id}/contest/sync")]
async fn sync_repository_contest(
    identity: Identity,
    pool: Data<DbPool>,
    path: Path<(i32,)>,
    request: HttpRequest,
) -> PageResult {
    use std::path::Path;

    let logged_user = require_identity(&identity)?;
    if !logged_user.is_admin {
        return Err(PageError::Forbidden(
            "Apenas administradores podem fazer isso".into(),
        ));
    }
    let (repository_id,) = path.into_inner();
    let mut connection = pool.get()?;

    let repository = repository::get_repository_by_id(&mut connection, repository_id)?;
    let git_repository =
        git2::Repository::open(Path::new(&format!("/data/{}/", repository.path))).unwrap();
    let head_commit = git_repository.head().unwrap();
    let head_commit_short_id = head_commit
        .peel_to_commit()
        .unwrap()
        .as_object()
        .short_id()
        .unwrap();
    let head_commit_shorthand = head_commit_short_id.as_str().unwrap();

    let contest_toml_contents =
        fs::read_to_string(format!("/data/{}/contest.toml", repository.path)).unwrap();
    let contest_config = toml::from_str::<ContestConfig>(&contest_toml_contents).unwrap();
    let contest = match contest::get_contest_by_name(&mut connection, &(String::from(&contest_config.br.name) + " (revisado)")) {
        Ok(contest) => contest,
        Err(NotFound) => contest::insert_contest(
            &mut connection,
            NewContest {
                name: contest_config.br.name,
                start_instant: None,
                end_instant: None,
                creation_user_id: logged_user.id,
                creation_instant: Local::now().naive_utc(),
                grade_ratio: None,
                grade_after_ratio: None,
            },
        )?,
        Err(e) => return Err(PageError::Database(e)),
    };
    let contest_problems = contest::get_contest_problems_by_contest(&mut connection, contest.id)?;
    for problem in contest_config.problems {
        let problem_id_prefix = format!("jughisto+{}+{}+", repository.path, problem.path);
        let problem_id = format!("{}{}", problem_id_prefix, head_commit_shorthand);
        let package_path = format!("/data/{}", problem_id);
        let repository_problem_path = format!("/data/{}/{}", repository.path, problem.path);
        let output = Command::new(format!("/data/{}/blem/blem", repository.path))
            .args([
                "make-full-jughisto-package",
                &package_path,
                &repository_problem_path,
            ])
            .output()
            .expect("failed to execute process");
        print!("{}", String::from_utf8_lossy(&output.stdout));
        print!("{}", String::from_utf8_lossy(&output.stderr));
        if output.status.success() {
            println!("Upserting problem");
            let blem_config = toml::from_str::<BlemConfig>(&fs::read_to_string(format!(
                "{}/blem.toml",
                package_path
            ))?)
            .unwrap();
            problem::upsert_problem(
                &mut connection,
                NewProblem {
                    id: problem_id.clone(),
                    name: problem.name,
                    memory_limit_bytes: blem_config.memory_limit_kb.unwrap_or(204800) * 1024,
                    time_limit_ms: blem_config.time_limit_ms,
                    status: "compiled".into(),
                    creation_instant: Local::now().naive_utc(),
                    creation_user_id: logged_user.id,
                },
            )?;
        }
        if let Some(contest_problem) = contest_problems
            .iter()
            .find(|x| x.problem_id.starts_with(&problem_id_prefix))
        {
            contest::update_contest_problem(
                &mut connection,
                contest_problem.id,
                problem_id.clone(),
                problem.label,
            )?;
        } else {
            contest::relate_problem(
                &mut connection,
                NewContestProblems {
                    label: problem.label,
                    contest_id: contest.id,
                    problem_id: problem_id.clone(),
                },
            )?;
        }
    }

    Ok(redirect_to_referer(
        format!(
            "Competição do repositório {} atualizada com sucesso!",
            repository.name
        ),
        &request,
    ))
}
